from cndi.annotations.component import ComponentClass
import logging

logger = logging.getLogger("cndi.annotations")

beans = list()
autowires = list()
components = list()
beanStore = dict()
componentStore = dict()
profiles = dict()

from functools import wraps
import importlib
import copy


def importModuleName(fullname):
    modules = fullname.split('.')
    module = importlib.import_module(modules[-1], package='.'.join(modules[:-1]))
    return module

def normaliseModuleAndClassName(name):
    nameList:list = name.split(".")
    if "__init__" in nameList:
        nameList.remove("__init__")
    return '.'.join(nameList)
def getBeanObject(objectType):
    objectType = normaliseModuleAndClassName(objectType)
    bean = beanStore[objectType]
    objectInstance = bean['objectInstance']
    # if (objectInstance.__class__.__name__ == "function"):
    #     args[key] = objectInstance()
    # else:
    return copy.deepcopy(objectInstance) if bean['newInstance'] else objectInstance


class AutowiredClass:
    def __init__(self, required, func, kwargs: dict()):
        self.fullname = '.'.join([func.__qualname__])
        self.className = normaliseModuleAndClassName('.'.join(func.__qualname__.split(".")[:-1]))
        self.func = func
        self.kwargs = kwargs
        self.required = required

    def dependencyInject(self):
        dependencies = self.calculateDependencies()
        dependencyNotFound = list()
        for dependency in dependencies:
            if dependency not in beanStore:
                dependencyNotFound.append(dependency)

        if len(dependencyNotFound) > 0:
            logger.warning(f"Skipping {self.fullname}")
            assert not self.required, "Could not initialize " + self.fullname + " with beans " + str(
                dependencyNotFound)

        kwargs = self.kwargs
        args = dict()
        for (key, value) in kwargs.items():
            fullName = normaliseModuleAndClassName('.'.join([value.__module__, value.__name__]))
            if fullName in beanStore:
                args[key] = getBeanObject(fullName)

        if self.className in beanStore:
            self.func(beanStore[self.className], **args)
        else:
            logger.info(f"{self.className} {self.fullname}")
            self.func(**args)

    def calculateDependencies(self):
        return list(
            map(lambda dependency: normaliseModuleAndClassName('.'.join([dependency.__module__, dependency.__name__])), self.kwargs.values()))


def Component(func: object):

    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    moduleName = wrapper.__module__[:-9] if wrapper.__module__.endswith(".__init__") else wrapper.__module__
    componentFullName = '.'.join([moduleName,wrapper.__qualname__])
    logger.info(f"Function name " + componentFullName)
    duplicateComponents = list(filter(lambda component: component.fullname == componentFullName, components))
    if duplicateComponents.__len__() > 0:
        logger.info(f"Duplicate Component found for: {duplicateComponents}")
    else:
        components.append(ComponentClass(**{
            'fullname': componentFullName,
            'func': wrapper,
            'annotations': wrapper.__init__.__annotations__ if "__annotations__" in dir(wrapper.__init__) else {}
        }))
    return  wrapper


def Bean(newInstance=False):
    def inner_function(func):
        annotations = func.__annotations__
        returnType = annotations['return']
        del annotations['return']

        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        annotations = dict(
            map(lambda key: (key, '.'.join([annotations[key].__module__, annotations[key].__qualname__])), annotations))
        beans.append({
            'name': '.'.join([returnType.__module__, returnType.__name__]),
            'newInstance': newInstance,
            'object': wrapper,
            'fullname': wrapper.__qualname__,
            'kwargs': annotations,
            'index': len(beans)
        })
        return wrapper

    return inner_function


def Autowired(required=True):
    def inner_function(func: object):
        annotations = func.__annotations__

        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        autowires.append(AutowiredClass(required=required, **{
            'kwargs': annotations,
            'func': wrapper
        }))

        return wrapper

    return inner_function

def getBean(beans, name):
    return list(filter(lambda x: x['name'] == name, beans))[0]


def workOrder(beans):
    allBeanNames = list(map(lambda bean: bean['name'], beans))
    beanQueue = list(filter(lambda bean: len(bean['kwargs']) == 0, beans))
    beanIndexes = list(map(lambda bean: bean['index'], beanQueue))

    beanDependents = list(filter(lambda bean: bean['index'] not in beanIndexes, beans))
    beanQueueNames = list(map(lambda bean: bean['name'], beanQueue))

    for i in range(len(beanQueue)):
        beanQueue[i]['index'] = i

    for dependents in beanDependents:
        args = list(dependents['kwargs'].values())
        flag = True
        for argClassName in args:
            if (argClassName not in beanQueueNames and argClassName in allBeanNames) or argClassName in beanQueueNames:
                flag = flag and True
                dependents['index'] = getBean(beans, argClassName)['index'] + max(beanIndexes)
            else:
                flag = False

        if flag:
            beanQueue.append(dependents)
            beanQueueNames.append(dependents['name'])

    assert len(beanQueue) == len(beans), "Somebeans were not initialized properly"
    return list(sorted(beanQueue, key=lambda x: x['index']))





import os
from pathlib import Path

import requests

if "ELFIE_DOWNLOAD_PATH" not in os.environ:
    path = os.path.join(Path.home(), ".elfie/")
    os.environ["ELFIE_DOWNLOAD_PATH"] = path
    if not os.path.exists(path):
        os.makedirs(path)

def getElfieDownloadPath():
    return os.environ["ELFIE_DOWNLOAD_PATH"]

class CDNFileDownloader(object):
    def __init__(self):
        self.url = "https://github.com/mayank31313/open_cdn/raw/main"
        self.downloadPath = getElfieDownloadPath()

    def downloadFile(self, childPath):
        with requests.get(f"{self.url}/{childPath}", stream=True) as r:
            r.raise_for_status()
            fullPath = os.path.join(self.downloadPath, childPath)
            if not os.path.exists(os.path.dirname(fullPath)):
                os.makedirs(os.path.dirname(fullPath))

            with open(fullPath, "wb") as file:
                for chunk in r.iter_content(chunk_size=8192):
                    file.write(chunk)

            return fullPath


import unittest

from cndi.annotations import Autowired
from cndi.initializers import AppInitilizer
from test_module.TestBean import TestBean


class AppInitializerTest(unittest.TestCase):
    def testComponentScanAndDI(self):
        @Autowired()
        def setTestBean(bean: TestBean):
            global testBean
            print(bean)
            testBean = bean

        app = AppInitilizer()
        app.componentScan("test_module")

        app.run()
        self.assertEqual(testBean.name, "Test 123")
        self.assertIsInstance(testBean, TestBean)


import os
import importlib
from importlib._bootstrap_external import _NamespacePath
import logging

logger = logging.getLogger("cndi.utils")

def injectEnvAsDict(dictObject, parent=""):
    for (key,value) in dictObject.items():
        if isinstance(value, dict):
            injectEnvAsDict(value, key + ".")
        elif isinstance(value, list) or isinstance(value, set) or isinstance(value, tuple):
             raise NotImplementedError("Value of type: %s object cannot be list, set or tuple" % (type(value)))
        else:
            injectEnvAsKeyValue(key, value)

def injectEnvAsKeyValue(key, value):
    if key not in os.environ:
        os.environ[key] = value

def dirModuleFilter(module, filt = lambda y: True):
    return filter(lambda x: not x.startswith('__') and filt(x), dir(module))

def walkDir(path):
    pythonFilesInComponent = list()
    for (root, dirs, files) in os.walk(path, topdown=path):
        pythonFilesInComponent.extend(map(lambda x: '.'.join([root[path.__len__():].replace('\\', '.'),x[:-3]]),filter(lambda x: x.endswith(".py"), files)))
    return pythonFilesInComponent

def walkChild(module):
    locations = module.__spec__.submodule_search_locations
    if isinstance(locations, _NamespacePath):
        l = module.__spec__.submodule_search_locations._path
    else:
        l = locations
    pythonComponents = list(map(lambda x: module.__name__ + x ,walkDir(l[0] if isinstance(l, list) else l)))
    return pythonComponents

def importSubModules(module, skipModules=[], callback=None):
    for m in walkChild(module):
        if len(list(filter(lambda x: m.startswith(x), skipModules))) > 0:
            logger.warning(f"Skipping ImportModule: {m}")
            continue
        logger.info(f"Importing: {m}")
        moduleInstance = importlib.import_module(m)
        if callback is not None:
            callback(moduleInstance)



import unittest

from cndi.annotations import Component, Bean, Autowired
from cndi.env import VARS
from cndi.initializers import AppInitilizer
from test_module.TestBean import TestBean


@Bean()
def getTestBean() -> TestBean:
    return TestBean("testBean")

@Component
class FirstComponent:
    def __init__(self):
        self.triggered = False
    def postConstruct(self):
        self.triggered = True

@Component
class SecondTestClass:
    def __init__(self, firstComponent: FirstComponent):
        self.firstComponent = firstComponent
        self.testBean = None

    def postConstruct(self, testBean: TestBean):
        self.testBean = testBean

class TestComponents(unittest.TestCase):
    def setUp(self) -> None:
        VARS.clear()
        self.store = dict()

    def testComponents(self):
        @Autowired()
        def setComponent(secondComponent: SecondTestClass, bean: TestBean):
            self.store['component'] = secondComponent
            self.store['bean'] = bean

        appInitializer = AppInitilizer()
        appInitializer.run()

        self.assertTrue(self.store['component'].firstComponent.triggered)
        self.assertIsNotNone(self.store['component'].testBean)


import copy
import importlib
import os

from cndi.binders.message import DefaultMessageBinder
import logging

from cndi.annotations import beanStore, workOrder, beans, components, componentStore, autowires, getBeanObject, getBean
from cndi.env import loadEnvFromFile, getContextEnvironment
from cndi.utils import importSubModules

logger = logging.getLogger(__name__)

class AppInitilizer:
    def __init__(self):
        self.componentsPath = list()
        applicationYml = "resources/application.yml"
        if os.path.exists(applicationYml):
            logger.info(f"External Configuration found: {applicationYml}")
            loadEnvFromFile(applicationYml)


    def componentScan(self, module):
        importModule = importlib.import_module(module)
        self.componentsPath.append(importModule)


    def run(self):
        for module in self.componentsPath:
            importSubModules(module)

        workOrderBeans = workOrder(beans)

        for bean in workOrderBeans:
            logger.info(f"Registering Bean {bean['fullname']}")
            kwargs = dict()
            for key, className in bean['kwargs'].items():
                tempBean = beanStore[className]
                kwargs[key] = copy.deepcopy(tempBean['object']) if tempBean['newInstance'] else tempBean['object']

            bean['objectInstance'] = bean['object'](**kwargs)
            beanStore[bean['name']] = bean

        for component in components:
            componentStore[component.fullname] = component
            kwargs = constructKeyWordArguments(component.annotations)
            objectInstance = component.func(**kwargs)
            if 'postConstruct' in dir(objectInstance):
                postConstructKArgs = constructKeyWordArguments(objectInstance.postConstruct.__annotations__)
                objectInstance.postConstruct(**postConstructKArgs)

            beanStore[component.fullname] = dict(objectInstance=objectInstance,
                                                 name=component.fullname,
                                                 object=objectInstance, index=0, newInstance=False,
                                                 fullname=component.func.__name__, kwargs=kwargs)

        messageBinderEnabled = getContextEnvironment("rcn.binders.message.enable", defaultValue=False, castFunc=bool)
        defaultMessageBinder = None

        if messageBinderEnabled:
            defaultMessageBinder = DefaultMessageBinder()
            defaultMessageBinder.performInjection()

        for autowire in autowires:
            autowire.dependencyInject()

        if defaultMessageBinder is not None:
            defaultMessageBinder.start()

def constructKeyWordArguments(annotations):
    kwargs = dict()
    for key, classObject in annotations.items():
        tempBean = beanStore[f"{classObject.__module__}.{classObject.__name__}"]
        kwargs[key] = copy.deepcopy(tempBean['object']) if tempBean['newInstance'] else tempBean['object']
    return kwargs



import threading
import time
import logging

from cndi.annotations import Component
from cndi.env import getContextEnvironment

class Event(object):
    def __init__(self, event_name=None, event_handler=None, event_object=None, event_invoker=None):
        self.event_name = event_name
        self.event_handler = event_handler
        self.event_invoker = event_invoker
        self.event_object = event_object

@Component
class EventHandler(threading.Thread):
    def __init__(self):
        threading.Thread.__init__(self)
        self.logger = logging.getLogger(f"{self.__class__.__module__}.{self.__class__.__name__}")
        self.EVENTS_MAP = dict()
        self.sleepwait = getContextEnvironment("rcn.events.waittime", 2.0, float)
        self.expectedInvokerTime = getContextEnvironment("rcn.events.expected.invoker.time", 0.003, float)
        self._enabled = getContextEnvironment("rcn.events.enable", defaultValue=False, castFunc=bool)
        self.logger.debug(f"Expected Invoker Time set to {self.expectedInvokerTime} seconds")
        self.logger.debug(f"Sleep time set to {self.sleepwait} seconds")

    def postConstruct(self):
        if self._enabled:
            self.start()

    def registerEvent(self, event: Event):
        self.EVENTS_MAP[event.event_name] = event

    def triggerEventExplicit(self, eventName, **kwargs):
        if eventName not in self.EVENTS_MAP or not self._enabled:
            return None

        eventObject:Event = self.EVENTS_MAP[eventName]
        eventObject.event_handler(kwargs, None)

    def run(self) -> None:
        while self._enabled:
            for event_name in self.EVENTS_MAP:
                event = self.EVENTS_MAP[event_name]
                try:
                    self.logger.debug(f"Calling event - {event_name}")
                    start = time.time()
                    callEvent = event.event_invoker(event.event_object)
                    timeDiff = time.time() - start
                    if timeDiff > self.expectedInvokerTime:
                        self.logger.warning(f"Time exceed for event invoker: {timeDiff} secs")
                    if callEvent is not None and 'trigger' in callEvent and callEvent['trigger']:
                        responseObject = event.event_handler(callEvent, event.event_object)
                        if responseObject is not None:
                            event.event_object = responseObject
                            self.EVENTS_MAP[event_name] = event

                except Exception as e:
                    self.logger.error(f"Exception occured while invoking event: {event_name} Exception: {e}")
            time.sleep(self.sleepwait)


import setuptools

VERSION = "2.1.6a"

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="cndi",
    version=VERSION,
    author="Mayank Shinde",
    author_email="mayank31313@gmail.com",
    description="",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


from threading import Thread

import pika
from pika import BlockingConnection
import logging

from pika.adapters.blocking_connection import BlockingChannel
from pika.spec import Basic

from cndi.binders.message import Message, SubscriberChannel
from cndi.binders.message.utils import extractChannelNameFromPropertyKey, MessageChannel
from cndi.env import getContextEnvironment, getContextEnvironments


class RabbitMQProducerBinding(MessageChannel):
    def __init__(self, connection: BlockingConnection):
        self.channel = connection.channel()
        self.topic = None

    def setTopic(self, topic):
        MessageChannel.setTopic(self, topic)
        self.channel.queue_declare(queue=self.topic)
        self.channel.exchange_declare(exchange=self.topic)
        self.channel.queue_bind(exchange=self.topic, queue=self.topic)

    def send(self, message: Message) -> None:
        self.channel.basic_publish(
            exchange=self.topic,
            routing_key=self.topic,
            body=message.message
        )

    def close(self):
        self.channel.close()

class RabbitMQSubscriberChannel(SubscriberChannel):
    def __init__(self, channel: BlockingChannel):
        SubscriberChannel.__init__(self)
        self.channel = channel

    def setTopic(self, topic, group="", channelName=None):
        SubscriberChannel.setTopic(self, topic)
        self.channel.queue_declare(queue=topic  + group)
        self.channel.queue_bind(exchange=topic, queue=topic + group, routing_key="#")



class RabbitMQBinder():
    def __init__(self):
        self.logger = logging.getLogger('.'.join([self.__class__.__module__, self.__class__.__name__]))
        brokerUrl = getContextEnvironment("rcn.binders.message.rabbitmq.brokerUrl", defaultValue=None)
        brokerPort = getContextEnvironment("rcn.binders.message.rabbitmq.brokerPort", defaultValue=None)

        brokerUrl = getContextEnvironment("rcn.binders.message.brokerUrl", required=True, defaultValue=brokerUrl)
        brokerPort = getContextEnvironment("rcn.binders.message.brokerPort", required=True, castFunc=int,
                                           defaultValue=brokerPort)

        contextEnvs = getContextEnvironments()

        self.connection = pika.BlockingConnection(pika.ConnectionParameters(host=brokerUrl, port=brokerPort))
        self.rabbotmqProducerChannelBindings = filter(lambda key: key.startswith('rcn.binders.message.rabbitmq.producer'),
                                                 contextEnvs)
        self.mqttConsumerChannelBindings = filter(
            lambda key: key.startswith('rcn.binders.message.rabbitmq.consumer') and key.endswith("destination"),
            contextEnvs)
        self.CHANNELS_TO_TOPIC_MAP = dict()
        self.channel = self.connection.channel()

        self.channelThread = Thread(target=self.channel.start_consuming)

    def stopConsumers(self):
        self.channel.stop_consuming()

    def bindSubscribers(self, CHANNELS_TO_FUNC_MAP):
        subscriptionTopics = list()
        topicConsumers = dict()
        binders = dict()

        def consumerMessage(ch: BlockingChannel, method: Basic.Deliver, properties, body):
            topicConsumers[method.exchange](body)

        for propertyKey in self.mqttConsumerChannelBindings:
            channelName = extractChannelNameFromPropertyKey(propertyKey)
            print(channelName)
            binderPath = f"rcn.binders.message.rabbitmq.consumer.{channelName}"
            consumerGroup = getContextEnvironment(f"{binderPath}.group")

            if channelName not in CHANNELS_TO_FUNC_MAP:
                self.logger.error(f"Channel not found: {channelName}")
                continue
            consumerBinding = RabbitMQSubscriberChannel(self.channel)
            callbackDetails = CHANNELS_TO_FUNC_MAP[channelName]
            callback = callbackDetails['func']
            consumerBinding.setOnConsumeCallback(callback)
            topicName = getContextEnvironment(propertyKey, required=True)
            subscriptionTopics.append(topicName)

            consumerBinding.setTopic(topicName, group=consumerGroup)
            if topicName in topicConsumers:
                raise KeyError(f"Duplicate topic found {topicName} with {topicConsumers[topicName]}")

            topicConsumers[topicName] = consumerBinding
            binders[channelName] = consumerBinding

            self.channel.basic_consume(queue=topicName + consumerGroup, auto_ack=True, on_message_callback=consumerMessage)

        return binders

    def bindProducers(self):
        binders = dict()
        for propertyKey in self.rabbotmqProducerChannelBindings:
            channelName = extractChannelNameFromPropertyKey(propertyKey)
            producerBinding = RabbitMQProducerBinding(connection=self.connection)

            topicName = getContextEnvironment(propertyKey, required=True)
            self.logger.info("Topic Name: "  + topicName)
            producerBinding.setTopic(topicName)
            self.CHANNELS_TO_TOPIC_MAP[channelName] = topicName
            binders[channelName] = producerBinding

        return binders